# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/general_messages.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass

import betterproto


class MessageLevel(betterproto.Enum):
    Message_LEVEL_ERROR = 0
    Message_LEVEL_WARNING = 1
    Message_LEVEL_INFO = 2


@dataclass(eq=False, repr=False)
class ErrorMessage(betterproto.Message):
    code: int = betterproto.int32_field(1)
    message: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class Matrix(betterproto.Message):
    v00: float = betterproto.double_field(1)
    v01: float = betterproto.double_field(2)
    v10: float = betterproto.double_field(3)
    v11: float = betterproto.double_field(4)
